/* Authors :
 Christian Lebiere, David Reitter, Ion Juvina, Andrea Stocco and Jasmeet Ajmani
 */

/* Desciption :
 * The bot consists of three interdependent components: action, prediction and selection.
 Those techniques are inspired by human cognition, and correspond to the functions of procedural strategies,
 declarative sequence learning, and metacognitive evaluation, respectively.  The workings of these components
 are detailed in the following sections.

Action
Action consists of a number of independent strategies.  The three default ones are, in order:

1.	Sticky. Pick a position at random, move there and stay there as long as the strategy is applied.
2.	OppW: Predict the next position of the weakest opponent (weakness is defined by least number of total points
    up to current time) then move opposite to it on the clock.  Each subsequent iteration in applying this strategy
    will require estimating the next position of that opponent and moving there.
3.	OppS: Same as the previous strategy (OppW) but applied to the stronger of the two opponents.

 Prediction
Action strategies 2 and 3 (OppW and OppS) require the ability to predict the next position of an opponent.  Each opponent
will be predicted separately.  The basic prediction algorithm that will be used is as follows:

1.	For every game update, record the position of each opponent at that iteration.
2.	Given the current opponent position ct and its previous position ct-1, compare it to all
    past position triplets cs, cs-1 and cs-2 for s<t.  If ct - ct-1 = cs-1 - cs-2 then make the
    prediction for that opponent ct+1 = ct + cs - cs-1.  That prediction carries a weight of 1/sqrt(t-s).
    If ct - ct-1 is not equal to cs-1 - cs-2 then no prediction is made for that value of s.
3.	Repeat for all past values of s<t, summing up the weights for all identical predictions (i.e., the same predicted value
    of ct+1).  The value of ct+1 with the highest weight is selected as the prediction for that opponent to be used in the
    action strategies above.  If no prediction can be made (i.e., there is no s<t for which ct - ct-1 = cs-1 - cs-2), then the
    opponent position is assumed to be the current one, i.e. ct+1 = ct.

 Selection
The action strategies will be picked in order, starting with Sticky, followed by OppW and OppS, and then back to Sticky.  The
current strategy will keep getting used as long as it is winning (defined as a payoff of 8 or more).  If it loses (i.e.,
gets a payoff of less than 8) two iterations in a row, then the next strategy in the order is selected.
*/

package strategy;

import game.Game;
import game.Strategy;
import game.Utility;
import java.util.Random;
import lemonade.LemonadeUtility;
import static java.lang.Math.pow;

class DataStorage{
    public static int count = 0;
    public static int action0[] = new int [100];
    public static int action1[] = new int [100];
    public static int action2[] = new int [100];
    public static int utility0[] = new int [100];
    public static int utility1[] = new int [100];
    public static int utility2[] = new int [100];
}

public class ACTR implements Strategy{

    int botaction;
    Utility utility;
    long randomseed;
    int numActions;
    int i;
    int scheme; /* Scheme = 0 => Sticky, Scheme = 1 => OppW, Scheme = 2 => OppS */
    
public ACTR (Game g, long seed, String options) {
      numActions = g.getNumActions();
      randomseed = seed;
      utility = g.getUtility();
}

  public int modulusfunc(int value)
  {   int num = 0;
      switch (value)
      {
          case 12: num =0; break;
          case 13: num =1; break;
          case 14: num =2; break;
          case 15: num =3; break;
          case 16: num =4; break;
          case 17: num =5; break;
          case 18: num =6; break;
          case 19: num =7; break;
          case 20: num =8; break;
          case 21: num =9; break;
          case 22: num =10; break;
          case -1: num =11; break;
          case -2: num =10; break;
          case -3: num =9; break;
          case -4: num =8; break;
          case -5: num =7; break;
          case -6: num =6; break;
          case -7: num =5; break;
          case -8: num =4; break;
          case -9: num =3; break;
          case -10: num =2; break;
          case -11: num =1; break;
      }
      return num;
  }
  /* Implementation of sticky actions */
  public void sticky(){
      Random r = new Random(randomseed);
      botaction = r.nextInt(numActions);
      scheme = 0;
  }

  /* Implemenatation of OppW actions */
  public void oppw(){
  int s;
  int j;
  int weakopponentposition = 0;
  int t;
  int utilitysum1 = 0;
  int utilitysum2 = 0;

  for (t=DataStorage.count -1;t>=0;t--)
  {
      utilitysum1 += DataStorage.utility1[t];
  }
  for (t=DataStorage.count -1;t>=0;t--)
  {
      utilitysum2 += DataStorage.utility2[t];
  }

  if (utilitysum1 < utilitysum2)
  {
  int temp,temp1,temp2;
  t= DataStorage.count -1;
  int currpos = DataStorage.action1[t];
  int prevpos = DataStorage.action1[t-1];
  int newpos[]= new int [t];
  double weight[]= new double [t];
  int matches[]=new int [t];
  double cumweight[]=new double[t];

  for(temp = t-1;temp>=0;temp--)
  {
      newpos[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      weight[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      matches[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      cumweight[temp]=0;
  }

  for (s=t-1;s>=2;s--)
  {
      temp1 = (DataStorage.action1[t]-DataStorage.action1[t-1]);
      temp2 = (DataStorage.action1[s-1]-DataStorage.action1[s-2]);
      if ((temp1 < 0) && (temp1 > -12))
          temp1 = modulusfunc(temp1);
      if ((temp2 < 0) && (temp2 > -12))
          temp2 = modulusfunc(temp2);
      if (temp1 == temp2)
      {
          temp = currpos + DataStorage.action1[s] - DataStorage.action1[s-1];
          if ((temp > 11)||((temp > -12)&& (temp <0)))
          {
              newpos[s]= modulusfunc(temp);
          }
          else
              newpos[s]= temp;
          weight[s] = 1/pow((t-s),0.5);
      }
  }

 for(s=t-1;s>=2;s--)
 {   
  for (j=t-1;j>=2;j--)
  {
    if ((newpos[s]==newpos[j])&&(weight[s]!=0)&&(weight[j]!=0)&&(s!=j))
    {
        matches[s]++;
    }
  }
 }
  for (j=t-1;j>=2;j--)
  {
      if (matches[j]>0)
      {
          cumweight[j]+=weight[j];
          for (s=t-1;s>=2;s--)
          {
              if ((newpos[j]==newpos[s])&& (s!=j))
                  cumweight[j]+=weight[s];
          }
      }
  }
  double maxcum = cumweight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (maxcum < cumweight[j])
      {
          maxcum = cumweight[j];
      }
  }
  double max = weight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (max < weight[j])
      {
          max = weight[j];
      }
  }
  if (maxcum < max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (max == weight[j])
          {
              weakopponentposition = newpos[j];
              break;
          }
      }
  }
  else if (maxcum > max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (maxcum == cumweight[j])
          {
              weakopponentposition = newpos[j];
              break;
          }
      }
  }
  else
      weakopponentposition = currpos;
  temp = weakopponentposition - 6;
  if ((temp > -7)&& (temp <0))
          {
              botaction = modulusfunc(temp);
          }
          else
              botaction = temp;
  }
  else
  {
  int temp,temp1,temp2;
  t = DataStorage.count -1;
  int currpos = DataStorage.action2[t];
  int prevpos = DataStorage.action2[t-1];
  int newpos[]= new int [t];
  double weight[]= new double [t];
  int matches[]=new int [t];
  double cumweight[]=new double[t];

  for(temp = t-1;temp>=0;temp--)
  {
      newpos[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      weight[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      matches[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      cumweight[temp]=0;
  }

  for (s=t-1;s>=2;s--)
  {
      temp1 = (DataStorage.action2[t]-DataStorage.action2[t-1]);
      temp2 = (DataStorage.action2[s-1]-DataStorage.action2[s-2]);
      if ((temp1 < 0) && (temp1 > -12))
          temp1 = modulusfunc(temp1);
      if ((temp2 < 0) && (temp2 > -12))
          temp2 = modulusfunc(temp2);
      if (temp1 == temp2)
      {
          temp = currpos + DataStorage.action2[s] - DataStorage.action2[s-1];
          if ((temp > 11)||((temp > -12)&& (temp <0)))
          {
              newpos[s]= modulusfunc(temp);
          }
          else
              newpos[s]= temp;
          weight[s] = 1/pow((t-s),0.5);
      }
      else
      {
          weight[s] = 0;
      }
  }

for(s=t-1;s>=2;s--)
 {
  for (j=t-1;j>=2;j--)
  {
    if ((newpos[s]==newpos[j])&&(weight[s]!=0)&&(weight[j]!=0)&&(s!=j))
    {
        matches[s]++;
    }
  }
 }
  for (j=t-1;j>=2;j--)
  {
      if (matches[j]>0)
      {
          cumweight[j]+=weight[j];
          for (s=t-1;s>=2;s--)
          {
              if ((newpos[j]==newpos[s])&& (s!=j))
                  cumweight[j]+=weight[s];
          }
      }
  }
  double maxcum = cumweight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (maxcum < cumweight[j])
      {
          maxcum = cumweight[j];
      }
  }
  double max = weight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (max < weight[j])
      {
          max = weight[j];
      }
  }
  if (maxcum < max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (max == weight[j])
          {
              weakopponentposition = newpos[j];
              break;
          }
      }
  }
  else if (maxcum > max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (maxcum == cumweight[j])
          {
              weakopponentposition = newpos[j];
              break;
          }
      }
  }
  else
      weakopponentposition = currpos;
  temp = weakopponentposition - 6;
  if ((temp > -7)&& (temp <0))
          {
              botaction = modulusfunc(temp);
          }
          else
              botaction = temp;
  }
  scheme = 1;
  }

  public void opps(){
  int s;
  int j;
  int strongopponentposition = 0;
  int t = DataStorage.count -1;
  int utilitysum1 = 0;
  int utilitysum2 = 0;

  for (t=DataStorage.count -1;t>=0;t--)
  {
      utilitysum1 += DataStorage.utility1[t];
  }
  for (t=DataStorage.count -1;t>=0;t--)
  {
      utilitysum2 += DataStorage.utility2[t];
  }

  if (utilitysum1 > utilitysum2)
  {
  int temp,temp1,temp2;
  t = DataStorage.count -1;
  int currpos = DataStorage.action1[t];
  int prevpos = DataStorage.action1[t-1];
  int newpos[]= new int [t];
  double weight[]= new double [t];
  int matches[]=new int [t];
  double cumweight[]=new double[t];

  for(temp = t-1;temp>=0;temp--)
  {
      newpos[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      weight[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      matches[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      cumweight[temp]=0;
  }

  for (s=t-1;s>=2;s--)
  {
      temp1 = (DataStorage.action1[t]-DataStorage.action1[t-1]);
      temp2 = (DataStorage.action1[s-1]-DataStorage.action1[s-2]);
      if ((temp1 < 0) && (temp1 > -12))
          temp1 = modulusfunc(temp1);
      if ((temp2 < 0) && (temp2 > -12))
          temp2 = modulusfunc(temp2);
      if (temp1 == temp2)
      {
          temp = currpos + DataStorage.action1[s] - DataStorage.action1[s-1];
          if ((temp > 11)||((temp > -12)&& (temp <0)))
          {
              newpos[s]= modulusfunc(temp);
          }
          else
              newpos[s]= temp;
          weight[s] = 1/pow((t-s),0.5);
      }
      else
      {
          weight[s] = 0;
      }
  }

for(s=t-1;s>=2;s--)
 {
  for (j=t-1;j>=2;j--)
  {
    if ((newpos[s]==newpos[j])&&(weight[s]!=0)&&(weight[j]!=0)&&(s!=j))
    {
        matches[s]++;
    }
  }
 }
  for (j=t-1;j>=2;j--)
  {
      if (matches[j]>0)
      {
          cumweight[j]+=weight[j];
          for (s=t-1;s>=2;s--)
          {
              if ((newpos[j]==newpos[s])&& (s!=j))
                  cumweight[j]+=weight[s];
          }
      }
  }
  double maxcum = cumweight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (maxcum < cumweight[j])
      {
          maxcum = cumweight[j];
      }
  }
  double max = weight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (max < weight[j])
      {
          max = weight[j];
      }
  }
  if (maxcum < max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (max == weight[j])
          {
              strongopponentposition = newpos[j];
              break;
          }
      }
  }
  else if (maxcum > max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (maxcum == cumweight[j])
          {
              strongopponentposition = newpos[j];
              break;
          }
      }
  }
  else
      strongopponentposition= currpos;
  temp = strongopponentposition - 6;
  if ((temp > -7)&& (temp <0))
          {
              botaction = modulusfunc(temp);
          }
          else
              botaction = temp;
  }
  else
  {
  int temp,temp1,temp2;
  t = DataStorage.count -1;
  int currpos = DataStorage.action2[t];
  int prevpos = DataStorage.action2[t-1];
  int newpos[]= new int [t];
  double weight[]= new double [t];
  int matches[]=new int [t];
  double cumweight[]=new double[t];

  for(temp = t-1;temp>=0;temp--)
  {
      newpos[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      weight[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      matches[temp]=0;
  }
  for(temp = t-1;temp>=0;temp--)
  {
      cumweight[temp]=0;
  }

  for (s=t-1;s>=2;s--)
  {
      temp1 = (DataStorage.action2[t]-DataStorage.action2[t-1]);
      temp2 = (DataStorage.action2[s-1]-DataStorage.action2[s-2]);
      if ((temp1 < 0) && (temp1 > -12))
          temp1 = modulusfunc(temp1);
      if ((temp2 < 0) && (temp2 > -12))
          temp2 = modulusfunc(temp2);
      if (temp1 == temp2)
      {
          temp = currpos + DataStorage.action2[s] - DataStorage.action2[s-1];
          if ((temp > 11)||((temp > -12)&& (temp <0)))
          {
              newpos[s]= modulusfunc(temp);
          }
          else
              newpos[s]= temp;
          weight[s] = 1/pow((t-s),0.5);
      }
      else
      {
          weight[s] = 0;
      }
  }

for(s=t-1;s>=2;s--)
 {
  for (j=t-1;j>=2;j--)
  {
    if ((newpos[s]==newpos[j])&&(weight[s]!=0)&&(weight[j]!=0)&&(s!=j))
    {
        matches[s]++;
    }
  }
 }
  for (j=t-1;j>=2;j--)
  {
      if (matches[j]>0)
      {
          cumweight[j]+=weight[j];
          for (s=t-1;s>=2;s--)
          {
              if ((newpos[j]==newpos[s])&& (s!=j))
                  cumweight[j]+=weight[s];
          }
      }
  }
  double maxcum = cumweight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (maxcum < cumweight[j])
      {
          maxcum = cumweight[j];
      }
  }
  double max = weight[t-1];
  for (j=t-2;j>=2;j--)
  {
      if (max < weight[j])
      {
          max = weight[j];
      }
  }
  if (maxcum < max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (max == weight[j])
          {
              strongopponentposition = newpos[j];
              break;
          }
      }
  }
  else if (maxcum > max)
  {
      for (j=t-1;j>=2;j--)
      {
          if (maxcum == cumweight[j])
          {
              strongopponentposition = newpos[j];
              break;
          }
      }
  }
  else
  strongopponentposition= currpos;
  temp = strongopponentposition - 6;
  if ((temp > -7)&& (temp <0))
          {
              botaction = modulusfunc(temp);
          }
          else
              botaction = temp;
  }
  scheme = 2;
  }

  @Override
  public int getAction() {
    if (DataStorage.count == 100)
        DataStorage.count = 0;
    else if (DataStorage.count == 0)
        sticky();
    else if (DataStorage.count >= 4)
          {
              if (DataStorage.utility0[DataStorage.count-1]<8 && DataStorage.utility0[DataStorage.count-2]<8)
              {
                  switch(scheme){
                          case 0:    oppw();
                                     break;
                          case 1:    opps();
                                     break;
                          case 2:    sticky();
                                     break;
              }
              }
              else
              {
                switch(scheme){
                         case 0:    sticky();
                                    break;
                         case 1:    oppw();
                                    break;
                         case 2:    opps();
                                    break;
                  }
              }
          }
          DataStorage.count ++;
    return botaction;
  }

  @Override
  public void observeOutcome(int[] actions) {
      DataStorage.action0[DataStorage.count-1]=actions[0];
      DataStorage.action1[DataStorage.count-1]=actions[1];
      DataStorage.action2[DataStorage.count-1]=actions[2];
      DataStorage.utility0[DataStorage.count-1]=utility.getUtility(actions)[0];
      DataStorage.utility1[DataStorage.count-1]=utility.getUtility(actions)[1];
      DataStorage.utility2[DataStorage.count-1]=utility.getUtility(actions)[2];
  }
}
