/* 
 * This class implements an agent that plays "The Lemonade Game".  The agent's
 * basic strategy is to establish a mutually beneficial cooperative relationship 
 * with one of the two other players, at the expense of the 3rd player.  
 * 
 * The agent has 3 main goals:
 * 1: Establish cooperation quickly (before both opponents decide to cooperate 
 *    with each other).
 * 2: Defend a cooperative position from the 3rd player.
 * 3: Break cooperation between opponents if they start cooperative relationship.
 * 
 * @author Amy Greenwald, Tom Goff, and Jordan Berg
 * @version 11/30/2010
 */

package edu.brown.cs.strategy;

import game.Game;
import game.Strategy;

import java.util.Vector;
import java.lang.Math;
import java.util.*;

public class GoffBot implements Strategy{
  Random rand;
  String state;
  GoffBot[] opponents;
  Vector<Integer> moveHist;//tracks past moves for the player = new Vector();
  Vector<Integer> scoreHist;//tracks history of player scores in each round
  
  int _thresh6att;
  int _thresh7att;
  int _threshPlayAcross;
  
  //Constructor
  public GoffBot(Game g, long seed, String options){
    this.state = "observe";//initial state is to make a random move and observe opponents' first moves
    //this.numActions = g.getNumActions();
    rand = new Random(seed + 1884);
    this.opponents = new GoffBot[2];
    this.opponents[0] = new GoffBot();
    this.opponents[1] = new GoffBot();
    this.moveHist = new Vector<Integer>();
    this.scoreHist = new Vector<Integer>();
    
    _thresh6att = 10;
    _thresh7att = 10;
    _threshPlayAcross = 9;
  }
  
  // This is only used to make models of opponents for tracking purposes
  public GoffBot(){
    this.moveHist = new Vector<Integer>();
    this.scoreHist = new Vector<Integer>();
  }
  
  
  //Given a location, return the index which is opposite that location
  static public int getOppositeSide(int location){
    while(location > 11) //make sure location is legitimate space on game field
      location -= 12;
    while(location < 0)
      location += 12;
    
    int opp = location + 6;
    while(opp > 11)
      opp -= 12;
    return opp;
  }
  
  //Given two locations, return the shortest distance between them on the 12-spotted board
  static public int dist(int a, int b){
    int big = b, small = a;
    if(a > b){
      big = a;
      small = b;
    }
    return Math.min( big - small, 12 - big + small );
  }
  
  //return the total score for the player
  public int getTotalScore(){
    int totalScore = 0;
    for(int i=0; i < this.scoreHist.size(); i++)
      totalScore += (this.scoreHist.get(i)).intValue();
    return totalScore;
  }
  
  //return true if an opponent is consistently choosing the same space at you are (attack that gives you 6 points)
  public boolean is6Attacked(){
    boolean atkA = true;
    int thresh = _thresh6att;
    for(int i=0; i < Math.min(thresh, this.moveHist.size()); i++)
      if( (moveHist.get(moveHist.size()-i-1)) - opponents[0].moveHist.get(moveHist.size()-i-1) != 0 )
        atkA = false; //only count attacks that last for N turns
    boolean atkB = true;
    for(int i=0; i < Math.min(thresh, this.moveHist.size()); i++)
      if( (moveHist.get(moveHist.size()-i-1)) - opponents[1].moveHist.get(moveHist.size()-i-1) != 0 )
        atkB = false; //only count attacks that last for N turns
    return (atkA || atkB) && (rand.nextInt(10) < 3);//when attack is realized, only react with some probability;
  }
  
  
  //return true if an opponent is consistently choosing a space next to you (attack that gives you 7 points)
  public boolean is7Attacked(){
    boolean atkA = true;
    int thresh = _thresh7att;
    for(int i=0; i < Math.min(thresh, this.moveHist.size()); i++)
      if( dist(moveHist.get(moveHist.size()-i-1), opponents[0].moveHist.get(moveHist.size()-i-1)) != 1 )
        atkA = false; //only count attacks that last for N turns
    boolean atkB = true;
    for(int i=0; i < Math.min(thresh, this.moveHist.size()); i++)
      if( dist(moveHist.get(moveHist.size()-i-1), opponents[1].moveHist.get(moveHist.size()-i-1)) != 1 )
        atkB = false; //only count attacks that last for N turns
    return (atkA || atkB) && (rand.nextInt(10) < 3);//when attack is realized, only react with some probability;
  }
  
  //true if attempts to play across from other players, "follow", isn't working
  public boolean playAcrossFail(){
    int thresh = _threshPlayAcross;
    if( this.moveHist.size() <=  thresh) {
      return false; //don't count very start
    }
    
    for(int i=0; i < thresh; i++){ //has playing Across worked at least once?
      if( getOppositeSide(moveHist.get(moveHist.size()-i-1)) == opponents[0].moveHist.get(moveHist.size()-i-1) ||
         getOppositeSide(moveHist.get(moveHist.size()-i-1)) == opponents[1].moveHist.get(moveHist.size()-i-1) )
        return false; //"Across" worked, any good opponent will cooperate
    }
    
    //if an opponent is holding still, Across will work
    if( opponents[0].moveHist.get(moveHist.size()-2) == opponents[0].moveHist.get(moveHist.size()-1) ||
       opponents[1].moveHist.get(moveHist.size()-2) == opponents[1].moveHist.get(moveHist.size()-1) )
      return false; //"Across" worked, any good opponent will cooperate
    
    //make sure not to stand still if that isn't working
    if( moveHist.get(moveHist.size()-1) == moveHist.get(moveHist.size()-2) &&
       moveHist.get(moveHist.size()-1) == moveHist.get(moveHist.size()-3) &&
       (scoreHist.get(moveHist.size()-1).intValue()+
        scoreHist.get(moveHist.size()-2).intValue()+
        scoreHist.get(moveHist.size()-3).intValue())/3.0 < 8) {
         return false; //
       }
    
    return true;//when attack is realized, only react with some probability;
  }
  
  
  
  public int getAction() {
    if( this.moveHist.size() < 1 ) //first move.  Make random move since no info available
      return rand.nextInt(12);//generates 0-11
    if ( this.moveHist.size() == 1 )
      return this.moveHist.lastElement().intValue();
    if ( this.moveHist.size() == 2 ){ //pick an opponent, with lowest score, with which to cooperate
      if(this.opponents[0].getTotalScore() >  this.opponents[1].getTotalScore() ){
        this.state = "coopB";
        return getOppositeSide(this.opponents[1].moveHist.lastElement());
      }
      this.state = "coopA";
      return getOppositeSide(this.opponents[0].moveHist.lastElement());
    }
    if( this.playAcrossFail() )//tried to play across from opponents, but they keep moving
      return this.moveHist.get(this.moveHist.size()-1);//try to "lead" by sticking still
    
    if( this.state.equals("coopA") ){
      //if not across from intended opponent... switch
      if( this.moveHist.lastElement() != getOppositeSide(this.opponents[0].moveHist.lastElement()) ){
        this.state = "coopB"; //else... "coopB"
        return getOppositeSide(this.opponents[1].moveHist.lastElement());
      }
      //if you are being attacked (3rd player sitting on top of or next to you)
      if( this.is6Attacked() || this.is7Attacked()){
        this.state = "coopB"; //switch to cooperating with B
        return getOppositeSide(this.opponents[1].moveHist.lastElement());
      }
      return getOppositeSide(this.opponents[0].moveHist.lastElement());
    }//end "coopA"
    
    else if( this.state.equals("coopB") ){
      //if not across from intended opponent... switch
      if( this.moveHist.lastElement() != getOppositeSide(this.opponents[1].moveHist.lastElement()) ){
        this.state = "coopA"; //else... "coopA"
        return getOppositeSide(this.opponents[0].moveHist.lastElement());
      }
      //if you are being attacked (3rd player sitting on top of or next to you)
      if( this.is6Attacked() || this.is7Attacked()){
        this.state = "coopA"; //switch to cooperating with B
        return getOppositeSide(this.opponents[0].moveHist.lastElement());
      }
      return getOppositeSide(this.opponents[1].moveHist.lastElement());
    }//end "coopB"
    return 8;//System.out.println("WARNING!  getAction() reached an invalid state");
  }
  
  
  
  public void observeOutcome(int[] actions) {
    int[] scores = scoreMoves(actions);
    
    this.moveHist.add( new Integer(actions[0]) );
    this.scoreHist.add( new Integer(scores[0]) );
    
    this.opponents[0].moveHist.add( new Integer(actions[1]) );//opponent A
    this.opponents[0].scoreHist.add( new Integer(scores[1]) );
    
    this.opponents[1].moveHist.add( new Integer(actions[2]) );//opponent B
    this.opponents[1].scoreHist.add( new Integer(scores[2]) );
    
    /*System.out.println("TJ move: "+actions[0]+" , state: "+this.state);
     System.out.println("TJ Score: "+scores[0]);
     System.out.println("A move: "+actions[1]);
     System.out.println("A Score: "+scores[1]);
     System.out.println("B move: "+actions[2]);
     System.out.println("B Score: "+scores[2]+"\n"); // */
  }
  
  //Given three moves, return the scores for each agent
  static public int[] scoreMoves(int[] moves){
    int[] scores = {0, 0, 0}; //scores for move1, move2, move3 respectively
    for(int p = 0; p < 3; p++) //for each GoffBot...
    {
      int score = 0;
      int loc = moves[p];  //get left score to the left
      do{
        score++;
        loc++;
        while(loc > 11)
          loc -= 12;
      }while(loc != moves[p] && loc != moves[(p+1)%3] && loc != moves[(p+2)%3]);
      
      loc = moves[p];  //get left score to the right
      do{
        score++;
        loc--;
        while(loc < 0)
          loc += 12;
      }while(loc != moves[p] && loc != moves[(p+1)%3] && loc != moves[(p+2)%3]);
      
      //divide score by number of people on same location
      int numShare = 0;
      for(int i=0; i<3; i++)
        if(moves[p] == moves[i])
          numShare++; //a GoffBot should "collide with itself",so at least 1
         scores[p] = score / numShare;
      }
      return scores;
   }

}//end GoffBot