package net;

import game.Game;

import java.io.IOException;
import java.net.SocketException;

import lemonade.ScaledLemonadeUtility;

/**
 * This is a unit test for the server for human players.
 * 
 * It tests:
 * 
 * 1. Running a full game
 * 2. A game aborted halfway (with closed sockets)
 * 3. A game with timeouts (no closed sockets)
 *
 * @author maz
 *
 */
public class TestRoom {

  public static Thread t;
  
  public static Room createRoom() throws IOException, InterruptedException{
    Game g= new Game(new int[]{12,12,12},new ScaledLemonadeUtility(12));
    int numGameServers = 10;
    int port = 5000;
    Room r = new Room(port, numGameServers, g, "",4,3000);
    t = new Thread(r);
    t.start();
    Thread.sleep(3000);
    return r;
  }
  
  
  public static void checkMessage(LineSocket socket,String expected) throws SocketException, IOException{
    String message = socket.receiveMessage();
    if (!message.equals(expected)){
      throw new RuntimeException("Expected "+expected+" Received "+message);
    }
  }
  
  public static void checkGameBegins(LineSocket A, LineSocket B, LineSocket C, int currentGameIndex) throws SocketException, IOException{
    checkMessage(A,"PING");
    checkMessage(B,"PING");
    checkMessage(C,"PING");
    A.sendMessage("ACK");
    B.sendMessage("ACK");
    C.sendMessage("ACK");
    String stringToSend="BEGIN:"+currentGameIndex+" LEMONADE 3 12";
    
    checkMessage(A,stringToSend+" 0");
    checkMessage(B,stringToSend+" 1");
    checkMessage(C,stringToSend+" 2");
  }
  
  
  
  public static void playGameWithAbort(int currentGameIndex) throws IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketB.openConnection();
    socketC.openConnection();
    
    checkGameBegins(socketA,socketB,socketC,currentGameIndex);
    
    for(int i=0;i<2;i++){
      socketA.sendMessage("0");
      socketB.sendMessage("1");
      socketC.sendMessage("2");
      checkMessage(socketA,"OUTCOME:0 1 2");
      checkMessage(socketB,"OUTCOME:1 0 2");
      checkMessage(socketC,"OUTCOME:2 1 0");

    }
    
    
    
    socketA.closeConnection();
    socketB.sendMessage("3");
    socketC.sendMessage("4");
    
    
    checkMessage(socketB,"ABORT");
    checkMessage(socketC,"ABORT");
    
    socketB.closeConnection();
    socketC.closeConnection();
  }
  
  public static void playFullGame(int currentGameIndex) throws SocketException, IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketB.openConnection();
    socketC.openConnection();

    checkGameBegins(socketA,socketB,socketC,currentGameIndex);
    
    socketA.sendMessage("CHAT:HELLO");
    checkMessage(socketA,"CHAT:HELLO");
    checkMessage(socketB,"CHAT:HELLO");
    checkMessage(socketC,"CHAT:HELLO");

    
    for(int i=0;i<3;i++){
      System.err.println("Iteration:"+i);
      socketA.sendMessage("0");
      socketB.sendMessage("1");
      socketC.sendMessage("2");
      checkMessage(socketA,"OUTCOME:0 1 2");
      checkMessage(socketB,"OUTCOME:1 0 2");
      checkMessage(socketC,"OUTCOME:2 1 0");
    }

    
    socketA.sendMessage("0");
    socketB.sendMessage("1");
    socketC.sendMessage("2");
    checkMessage(socketA,"GAMEOVER:0 1 2");
    checkMessage(socketB,"GAMEOVER:1 0 2");
    checkMessage(socketC,"GAMEOVER:2 1 0");

  }

  
  public static void connectThenSilentTimeout(int currentGameIndex) throws SocketException, IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketB.openConnection();
    socketC.openConnection();
    checkGameBegins(socketA,socketB,socketC,currentGameIndex);
    
    socketA = null;
    socketB = null;
    socketC.sendMessage("5");
    checkMessage(socketC,"ABORT");
  }

  
  
  public static void playerQuitsBeforeGameBegins(int currentGameIndex) throws SocketException, IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketA.closeConnection();
    
    socketB.openConnection();
    socketC.openConnection();

    String stringToSend="PING";
    checkMessage(socketB,stringToSend);
    checkMessage(socketC,stringToSend);
    
    
    
    stringToSend="ABORT";
    checkMessage(socketB,stringToSend);
    checkMessage(socketC,stringToSend);
    
    
  }


  public static void playThenSilentTimeout(int currentGameIndex) throws SocketException, IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketB.openConnection();
    socketC.openConnection();
    checkGameBegins(socketA,socketB,socketC,currentGameIndex);

    for(int i=0;i<2;i++){
      System.err.println("Iteration:"+i);
      socketA.sendMessage("0");
      socketB.sendMessage("1");
      socketC.sendMessage("2");
      checkMessage(socketA,"OUTCOME:0 1 2");
      checkMessage(socketB,"OUTCOME:1 0 2");
      checkMessage(socketC,"OUTCOME:2 1 0");
    }
    
    socketA = null;
    socketB = null;
    socketC.sendMessage("5");
    checkMessage(socketC,"ABORT");
  }


  public static void playerSends12(int currentGameIndex) throws SocketException, IOException{
    LineSocket socketA = new LineSocket();
    LineSocket socketB = new LineSocket();
    LineSocket socketC = new LineSocket();
    socketA.openConnection();
    socketB.openConnection();
    socketC.openConnection();
    checkGameBegins(socketA,socketB,socketC,currentGameIndex);

    for(int i=0;i<2;i++){
      System.err.println("Iteration:"+i);
      socketA.sendMessage("CHAT:HELLO");
      checkMessage(socketA,"CHAT:HELLO");
      checkMessage(socketB,"CHAT:HELLO");
      checkMessage(socketC,"CHAT:HELLO");
      socketA.sendMessage("0");
      socketB.sendMessage("1");
      socketC.sendMessage("2");
      checkMessage(socketA,"OUTCOME:0 1 2");
      checkMessage(socketB,"OUTCOME:1 0 2");
      checkMessage(socketC,"OUTCOME:2 1 0");
    }
    
    socketC.sendMessage("12");
    checkMessage(socketA,"ABORT");
    checkMessage(socketB,"ABORT");
    checkMessage(socketC,"ABORT");
  }

  public static void checkNumberOfPlayers(Room r, int number){
    int numUsersInSystem=r.getNumUsers();
    if (numUsersInSystem!=number){
      throw new RuntimeException("Wrong number of users in the system:"+numUsersInSystem);
      
    }
    
  }
  
  public static void main(String[] args) throws IOException, InterruptedException{
    Room r=null;
    long sleepTime=3000;
    try{
    r = createRoom();
    System.err.println("Beginning playGameWithAbort experiment");
    
    playGameWithAbort(1);
    System.err.println("Sleeping for ten seconds");
    Thread.sleep(sleepTime);
    checkNumberOfPlayers(r,0);
    
    playFullGame(2);
    
        
    
    System.err.println("Beginning connectThenSilentTimeout experiment");
    connectThenSilentTimeout(3);
    Thread.sleep(sleepTime);
    
    System.err.println("Beginning playThenSilentTimeout experiment");
    
    playThenSilentTimeout(4);
    Thread.sleep(sleepTime);
    checkNumberOfPlayers(r, 0);
    
    System.err.println("Beginning playerQuitsBeforeGameBegins experiment");
    
    playerQuitsBeforeGameBegins(5);
    Thread.sleep(sleepTime);
    checkNumberOfPlayers(r, 0);
    
    System.err.println("Beginning playerSends12 experiment");
    
    playerSends12(6);
    Thread.sleep(sleepTime);
    checkNumberOfPlayers(r, 0);
    
    System.err.println("Reached end of test");
    } catch (Exception e){
      e.printStackTrace();

    } finally {
      r.active = false;
      t.interrupt();
    }
    
  }
}
