package results;

import game.Game;
import game.Strategy;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import stats.SeriesObserver;
import stats.UtilitySeriesObserver;
import util.Array;
import lemonade.ScaledLemonadeUtility;


/**
 * Given a tournament with a larger number of series, this class
 * filters the tournaments to the output of a smaller number of series.
 * @author maz
 *
 */
public class SimulateLemonadeTournament {

  Hashtable<String,TripleResults> results;
  
  
  
  public SimulateLemonadeTournament(File dataFile, int repeats) throws IOException {
    Vector<TripleResults> resultsVector = TripleResults.getResultsWithVariance(dataFile, repeats);
    results = new Hashtable<String,TripleResults>();
    for(TripleResults result:resultsVector){
      String key = Array.arrayToNakedString(result.playerNames);
      if (results.containsKey(key)){
        throw new RuntimeException("SimulateLemonadeTournament cannot handle multiple series between the same bots in the same order");
      }
      results.put(key, result);
    }
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public void runTournament(String[] strategies){
    for(int i=0;i<strategies.length;i++){
      for(int j=i+1;j<strategies.length;j++){
        for(int k=j+1;k<strategies.length;k++){
          String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
          runGame(threeStrategies);
        }
      }
    }
  }

  /**
   * Run a tournament with an arbitrary game.
   * Every three players (including repeats) are played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public void runAsymmetricTournament(String[] strategies){
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        for(int k=0;k<strategies.length;k++){
          String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
          runGame(threeStrategies);
        }
      }
    }
  }
  
  /**
   * Run a tournament with an arbitrary game.
   * Every three players (not including repeats) are played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public void runAsymmetricNonSelfPlayTournament(String[] strategies){
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        for(int k=0;k<strategies.length;k++){
          if (i!=j && i!=k && j!=k){
            String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
            runGame(threeStrategies);
          }
        }
      }
    }
  }
  
  
  public TripleResults getTriple(String[] strategies){
    String key = Array.arrayToNakedString(strategies);
    return results.get(key);
  }
  
  public TripleResults mapTriple(String[] players, String[] strategies){
    TripleResults triple = getTriple(strategies);
    return new TripleResults(players,triple.scores,triple.squaredScores,triple.repeats);
  }
  
  
  /**
   * Simulate the competition, but map players to strategies.
   * Every three players (not including repeats) are played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public Vector<TripleResults> runAsymmetricNonSelfPlayTournament(String[] players, String[] strategies){
    Vector<TripleResults> newResults = new Vector<TripleResults>();
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        for(int k=0;k<strategies.length;k++){
          if (i!=j && i!=k && j!=k){
            String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
            String[] threePlayers = {players[i],players[j],players[k]};
            
            newResults.add(mapTriple(threePlayers,threeStrategies));
          }
        }
      }
    }
    return newResults;
  }
  
  

  
  public void runGame(String[] threeStrategies){
    TripleResults result = getTriple(threeStrategies);
    System.out.println(result);
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public void run2vs1Tournament(String[] strategies){
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        if (i!=j){
          String[] threeStrategies = {strategies[i],strategies[i],strategies[j]};
          runGame(threeStrategies);
        }
      }
    }
  }

  public static void showUsage(){
    System.err.println("Usage:SimulateLemonadeTournament [-data <DATAFILE>| -matchtype <MATCHTYPE>]* -players [<CLASS> <OPTION>]*");
    System.exit(0);
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   * @throws IOException 
   */
  public static void main(String[] args) throws IOException{
    
    if (args.length<4){
      showUsage();
    }
    
    int repeats = 1000;
    File dataFile = null;
    int playerArgIndex=args.length;
    String matchType = "normal";
    for(int i=0;i<args.length;i+=2){
      String arg = args[i];
      String fieldStr = args[i+1];
      if (arg.equals("-data")){
        dataFile = new File(fieldStr);
      } else if (arg.equals("-matchtype")){
        matchType = fieldStr;
      } else if (arg.equals("-repeats")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of repeats needs to be less than 2^31-1");
          showUsage();
        }
        repeats = (int)field;
      } else if (arg.equals("-players")){
        playerArgIndex=i+1;
        break;
      } else {
        System.err.println("Unknown option:"+arg);
        showUsage();
        
      }
    }
    if (dataFile==null){
      showUsage();
    }

    String[] strategies = new String[(args.length-playerArgIndex)/2];
    String[] options = new String[strategies.length];
    int playerIndex = 0;
    for(int i=playerArgIndex;i<args.length;i+=2){
      strategies[playerIndex]=args[i];
      options[playerIndex]=args[i+1];
      playerIndex++;
    }
    
    SimulateLemonadeTournament tournament = new SimulateLemonadeTournament(dataFile,repeats);
    if (matchType.equals("normal")){
      tournament.runTournament(strategies);
    } else if (matchType.equals("2vs1")){
      tournament.run2vs1Tournament(strategies);
    } else if (matchType.equals("all")){
      tournament.runTournament(strategies);
      tournament.run2vs1Tournament(strategies);
    } else if (matchType.equals("asymmetric")){
      tournament.runAsymmetricTournament(strategies);
    } else if (matchType.equals("asymmetricnsp")){
      tournament.runAsymmetricNonSelfPlayTournament(strategies);
    }
  }
}
