package stats.collusion;

import stats.SeriesObserver;
import game.Game;


/**
 * An observer of utility, like jointactionutility, that does
 * not keep track of joint actions, simply utilities.
 * @author maz
 *
 */
public abstract class AbstractCollusionObserver implements SeriesObserver{
  
  
  
  /**
   * The indices of the players currently in collusion.
   * "Pseudo-collusions", where two players are exactly opposite another,
   * can't initialize a collusion.
   */
  int[] collusionPlayers;
  
  /**
   * If collusion occurs:
   * 
   * collusionActions[collusionPlayers[0]]=actions[collusionPlayers[0]]
   * collusionActions[collusionPlayers[1]]=actions[collusionPlayers[1]]
   * 
   */
  int[] collusionActions;
  
  
  int[] actions;
  
  int roundCollusionStarted;

  int currentRound;
  
  protected boolean wasInCollusion(){
    return collusionPlayers!=null;
  }
  
  
  private boolean stillInCollusion(){
    return (collusionActions[collusionPlayers[0]]==actions[collusionPlayers[0]])
      && (collusionActions[collusionPlayers[1]]==actions[collusionPlayers[1]]);
  }
  
  private boolean isInCollusion(int actionA, int actionB){
    return ((actionA+ 6) % 12==actionB);
  }
  
  private void setCollusion(int playerA, int playerB){
    collusionPlayers = new int[2];
    collusionPlayers[0]=playerA;
    collusionPlayers[1]=playerB;
    collusionActions=new int[3];
    this.roundCollusionStarted = currentRound;
    System.arraycopy(actions,0,collusionActions,0,actions.length);
  }
  
  
  private boolean testPseudoCollusion(){
    int collusionCount = 0;
    for(int i=0;i<actions.length;i++){
      for(int j=i+1;j<actions.length;j++){
        if (isInCollusion(actions[i],actions[j])){
          collusionCount++;
        }
      }
    }
    return (collusionCount>=2);
  }
  
  
  /**
   * Called when a new collusion begins.
   * If a collusion is in progress, this is not called.
   */
  public abstract void observeStartCollusion();
  
  private void testBeginCollusion(){
    for(int i=0;i<actions.length;i++){
      for(int j=i+1;j<actions.length;j++){
        if (isInCollusion(actions[i],actions[j])){
          if (testPseudoCollusion()){
            return;
          }
          setCollusion(i,j);
          observeStartCollusion();
          return;
        }
      }
      
    }
  }
  
  
  
  public AbstractCollusionObserver(Game g){
    this();
  }
  /**
   * Construct a utility observer.
   * @param utility a utility function to be used.
   * @param numPlayers the number of players in the game.
   */
  public AbstractCollusionObserver(){
  }
  
  
  /**
   * The odd man out. If collusionPlayers = {0,1}, this is 2,
   * if collusioPlayers = {1,2}, this is 0, if collusionPlayers = {0,2}, this is 1.
   * @return
   */
  public int getSucker(){
    return 3-(collusionPlayers[0]+collusionPlayers[1]);
  }
  
  /**
   * Eliminate the records of the previous collusion.
   * This is not called until after the sub-class has
   * been notified that the collusion has been terminated.
   */
  private void terminateCollusion(){
    this.collusionActions = null;
    this.collusionPlayers = null;
  }
  
  
  /**
   * Called during the last round.
   * @param countBroken
   */
  public abstract void observeTerminateCollusion(boolean countBroken);
  
  
  
  /**
   * Called during a round of collusion that is neither the first nor the last.
   */
  public abstract void observeContinueCollusion();
  
  public static int MIN_LENGTH = 3;
  public void observeOutcome(int[] actions) {
    this.actions = actions;
    if (wasInCollusion()){
      if (stillInCollusion()){
        observeContinueCollusion();
      } else {
        observeTerminateCollusion(true);
        terminateCollusion();
        // A new collusion could begin, just as another ends.
        testBeginCollusion();
      }
    } else {
      testBeginCollusion();
    }
    
    this.currentRound++;
  }
  
  
  public int getCollusionLength(){
    return (this.currentRound - this.roundCollusionStarted);
  }
  
  public boolean collusionDidNotCount(){
    return (getCollusionLength()<=MIN_LENGTH);
  }
  
  

  public void finishGame() {
    if (wasInCollusion()){
      observeTerminateCollusion(false);
    }

    terminateCollusion();
  }

  public void finishSeries() {
  }
}
