/*
 * Solo Team Member: Richard Gibson
 * Institution: University of Alberta
 * Position: PhD Student
 *
 * NOTE: This program currently takes the full 10 seconds per 100 rounds!
 *
 * The UctBehaviourModeller strategy uses two components, an opponent modeller
 * and an optimization routine, to decide where to set up its lemonade stand on
 * each round.  First, the opponent modeller contains a small library of
 * simple behaviours, such as "playing constant" or "sitting across from my
 * previous location", that the modeller monitors each opponent doing independently.
 * Each time an opponent takes the same action that a behaviour suggests, that
 * behaviour "scores a point" for that particular opponent.  Note that
 * behaviour actions can overlap.  For example, if an opponent repeats his or her
 * previous action and in doing so plays across from my previous location, both
 * behaviours score a point.  In addition, each behaviour may have a
 * precondition that if not satisfied, the behaviour is not monitored for the
 * round.  An example of this is the "sit on me" behaviour,
 * which is only active after the opponent had a bad round (scored < 8 utility).
 * Furthermore, points decay over time so that the more recent rounds effect
 * the behaviour scores more than earlier rounds.
 *
 * Secondly, the optimization routine runs UCT (Bandit based Monte Carlo
 * planning, Kocsis and Szepesvari, 2006) with the opponent modeller to
 * find an "optimal" action for the current round.  For each opponent and each
 * behaviour, we have a "proportional score" equal to the behaviour's score
 * divided by the behaviour's maximum possible score (i.e. its score if it
 * had received a point on every round in which is was monitored).  For each
 * opponent, these proportional scores are summed and normalized, creating a
 * probability distribution over behaviours.  On each UCT simulation, the
 * opponents' actions are sampled from these two behaviour distributions while
 * our actions are sampled in the normal UCT way (argmax_a Q(s,a) + c*sqrt(log(n(s))/n(s,a))).
 * Simulations are run to a fixed horizon to speed up simulations; the horizon
 * is currently set to simulate just 3 rounds.  Simulations are run for equal
 * time lengths on each round, where the total time spent over 100 rounds is 
 * 10 seconds.  Finally, the action that eventually leads to the highest average
 * utility over the next 3 rounds is taken, as estimated by the UCT simulations.
 *
 * Below is a full list of the behaviours included in the opponent modeller:
 *
 * Name: Constant After Good
 * Precondition: Opponent's previous round was "good" (scored >= 8 utility)
 * Description: Returns the same action that the opponent took on the previous
 * round.
 *
 * Name: Constant After Bad
 * Precondition: Opponent's previous round was "bad" (scored < 8 utility)
 * Description: Returns the same action that the opponent took on the previous
 * round.
 *
 * Name: Across From Me
 * Precondition: None
 * Description: Returns the action opposite of my previous location.
 *
 * Name: Across From Other Opponent
 * Precondition: None
 * Description: Returns the action opposite of the other opponent's previous
 * location.
 *
 * Name: Across From Average
 * Precondition: None
 * Description: Returns the action that is oppositie of the "average" of my
 * previous location and the other opponent's previous location.  For example,
 * if I and the other opponent play 9 and 1 respectively, then the "average" is
 * 11, and so the action returned is 5 since it is across from 11.  When the
 * average is not an integer, the floor or ceiling is returned at random.
 *
 * Name: Best Overall
 * Precondition: None
 * Description: Returns the single action that in hindsight would have
 * gained the most utility over all previous rounds.  Ties are broken randomly.
 *
 * Name: Sandwich Me
 * Precondition: None
 * Description: Returns the action adjacent to my previous location that would
 * have given me the lowest utility on the previous round.
 *
 * Name: Sandwich Other Opponent
 * Precondition: None
 * Description: Returns the action adjacent to the other opponent's previous
 * location that would have given him or her the lowest utility on the
 * previous round.
 *
 * Name: Sit on Me
 * Precondition: Opponent's previous round was bad
 * Description: Returns my previous action.
 *
 * Name: Sit on Other Opponent
 * Precondition: Opponent's previous round was bad
 * Description: Returns the other opponent's previous action.
 *
 * Name: Match Me (or Mirror Me)
 * Precondition: Me and other opponent chose different actions on the previous
 * round.
 * Description: Returns the action not at my previous location that would have
 * given both me and the behaviour the same utility on the previous round.
 *
 * Name: Match Other Opponent (or Mirror Other Opponent)
 * Precondition: Me and other opponent chose different action on the previous
 * round.
 * Description: Returns the action not at the other opponent's previous
 * location that would have given both the other opponent and the behaviour
 * the same utility on the previous round.
 *
 * Name: Mystery
 * Precondition: None
 * Description: Returns any action that none of the above behaviours would
 * return.
 */

package edu.ualberta.rggibson;

import game.Game;
import game.Strategy;
import game.Utility;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.Random;

/**
 *
 * @author Richard G. Gibson
 */
public class UctBehaviourModeller implements Strategy
{
    final boolean DO_DEBUG = false;
    final int MAX_WALLTIME = 10000; // In milliseconds
    final int NUM_ITERATIONS = 100;
    final int NUM_PLAYERS = 3;

    // Parameters that could be tweaked (along with BehaviourModel's DECAY
    // parameter)
    final boolean DO_SAMPLE_BEHAVIOURS = true;
    final int MAX_DEPTH = 3;
    final double UCT_CONSTANT = 30;
    final double MOVING_THRESHOLD = 0.3333;

    int numActions;
    Utility utilFunc;
    Random rand;
    int time;
    long walltimePassed;

    int[] prevActions;
    double[][] nextActionNums;
    double[][] nextActionDenoms;
    double[] nextActionProbsSum;
    double[] nextBehaviourProbsSum;

    ArrayList<BehaviourModel>[] bms;

    public UctBehaviourModeller(Game g, long seed, String options)
    {
        long getConstructionStartTime = System.currentTimeMillis();

        numActions = g.getNumActions();
        utilFunc = g.getUtility();
        rand = new Random(seed);
        time = 0;

        prevActions = new int[NUM_PLAYERS];

        if (DO_SAMPLE_BEHAVIOURS)
        {
            nextBehaviourProbsSum = new double[2];
        }
        else
        {
            nextActionNums = new double[2][numActions];
            nextActionDenoms = new double[2][numActions];
            nextActionProbsSum = new double[2];
        }

        bms = new ArrayList[NUM_PLAYERS - 1];
        for (int p = 0; p < NUM_PLAYERS - 1; ++p)
        {
            bms[p] = new ArrayList<BehaviourModel>();
            bms[p].add(new BehaviourConstantAfter(p + 1, utilFunc, true));
            bms[p].add(new BehaviourConstantAfter(p + 1, utilFunc, false));
            bms[p].add(new BehaviourAcross(numActions, p + 1, -1));
            bms[p].add(new BehaviourAcross(numActions, p + 1, 0));
            bms[p].add(new BehaviourAcross(numActions, p + 1, 2 - p));
            bms[p].add(new BehaviourBestOverall(numActions, p + 1, utilFunc));
            bms[p].add(new BehaviourSandwich(numActions, p + 1, 0, utilFunc));
            bms[p].add(new BehaviourSandwich(numActions, p + 1, 2 - p, utilFunc));
            bms[p].add(new BehaviourSitOnAfterBad(p + 1, 0, utilFunc));
            bms[p].add(new BehaviourSitOnAfterBad(p + 1, 2 - p, utilFunc));
            bms[p].add(new BehaviourMatching(numActions, p + 1, 0));
            bms[p].add(new BehaviourMatching(numActions, p + 1, 2 - p));
            bms[p].add(new BehaviourMystery(p + 1, numActions, bms[p]));
        }

        walltimePassed = System.currentTimeMillis() - getConstructionStartTime;
    }
    
    public static int mod(int num, int modulus)
    {
        int result = num % modulus;
        if (result < 0)
        {
            result += modulus;
        }

        return result;
    }

    @Override
    public int getAction()
    {
        if (time == 0)
        {
            return rand.nextInt(numActions);
        }
        if (time == 1)
        {
            // Sit the furthest away from the opponents' previous locations
            int diff = Math.abs(prevActions[1] - prevActions[2]);
            int smallerLoc = ((prevActions[1] <= prevActions[2]) ? prevActions[1] : prevActions[2]);
            int action = smallerLoc + (diff / 2);
            if (diff < (numActions / 2))
            {
                action += (numActions / 2);
            }
            if (mod(diff, 2) == 1)
            {
                action += rand.nextInt(2);
            }

            return mod(action, numActions);
        }

        long getActionStartTime = System.currentTimeMillis();

        // Start running UCT simulations
        int maxDepth = (int)Math.min(NUM_ITERATIONS - time, MAX_DEPTH);
        double[][] stateValues = new double[maxDepth][];
        int[][] numVisits = new int[maxDepth][];
        for (int turn = 0; turn < maxDepth; ++turn)
        {
            stateValues[turn] = new double[(int)Math.pow(numActions, turn + 1)];
            numVisits[turn] = new int[(int)Math.pow(numActions, turn + 1)];
        }
        long maxSimulationTime = (MAX_WALLTIME - walltimePassed - 1) / (NUM_ITERATIONS - time);
        int sim = 0;
        while ((System.currentTimeMillis() - getActionStartTime) < maxSimulationTime)
        {
            // Determine the actions to simulate for us
            int[][] simulatedActions = new int[maxDepth][NUM_PLAYERS];
            for (int turn = 0; turn < maxDepth; ++turn)
            {
                // Find the best looking action: value + exploration
                ArrayList<Integer> bestActions = new ArrayList<Integer>();
                double bestActionValue = -Double.MAX_VALUE;
                for (int a = 0; a < numActions; ++a)
                {
                    // Base numActions representation for turn's stateIndex
                    int stateIndex = a;
                    for (int i = 0; i < turn; ++i)
                    {
                        stateIndex += (int)(Math.pow(numActions, i + 1)) * simulatedActions[i][0];
                    }
                    int numStateVisits;
                    if (turn > 0)
                    {
                        int oldStateIndex = simulatedActions[turn - 1][0];
                        for (int i = 0; i < turn - 1; ++i)
                        {
                            oldStateIndex += (int)(Math.pow(numActions, i + 1)) * simulatedActions[i][0];
                        }
                        numStateVisits = numVisits[turn - 1][oldStateIndex];
                    }
                    else
                    {
                        numStateVisits = sim;
                    }
                    double value;
                    if (numVisits[turn][stateIndex] > 0)
                    {
                        value = stateValues[turn][stateIndex] + UCT_CONSTANT * Math.sqrt(Math.log(numStateVisits * 1.0 / numVisits[turn][stateIndex]));
                    }
                    else
                    {
                        value = Double.MAX_VALUE;
                    }

                    if (value > bestActionValue)
                    {
                        bestActions.clear();
                        bestActions.add(a);
                        bestActionValue = value;
                    }
                    else if (value == bestActionValue)
                    {
                        bestActions.add(a);
                    }
                }

                // Randomly choose among the best actions to simulate
                int size = bestActions.size();
                assert(size > 0);
                int index = rand.nextInt(size);
                simulatedActions[turn][0] = bestActions.get(index);
            }

            // Evaluate how well those actions "will" do

            // Need to clone the models so that we can query and update them without
            // altering their member values
            int util = 0;
            BehaviourModel[][] uctBms = new BehaviourModel[2][bms[0].size()];
            for (int opp = 0; opp < 2; ++opp)
            {
                for (int m = 0; m < uctBms[opp].length; ++m)
                {
                    uctBms[opp][m] = bms[opp].get(m).clone();
                }
                
                // HACK: Set the behaviour mystery model's models
                BehaviourMystery myst = (BehaviourMystery)(uctBms[opp][uctBms[opp].length - 1]);
                myst.setBehaviourModels(uctBms[opp], uctBms[opp].length - 1);
            }

            for (int turn = 0; turn < maxDepth; ++turn)
            {
                if (DO_SAMPLE_BEHAVIOURS)
                {
                    for (int opp = 0; opp < 2; ++opp)
                    {
                        // First, sample a behaviour
                        double sum = 0.0;
                        if (turn == 0)
                        {
                            // Just grab the stored value for the first turn
                            sum = nextBehaviourProbsSum[opp];
                        }
                        else
                        {
                            for (int m = 0; m < uctBms[opp].length; ++m)
                            {
                                if (uctBms[opp][m].isAvailable())
                                {
                                    sum += uctBms[opp][m].getWeight();
                                }
                            }
                        }
                        int m;
                        if (sum > 0)
                        {
                            double dart = rand.nextDouble() * sum;
                            double cumProb = 0.0;
                            for (m = 0; m < uctBms[opp].length; ++m)
                            {
                                cumProb += uctBms[opp][m].getWeight();
                                if (dart < cumProb)
                                {
                                    break;
                                }
                                assert(m < uctBms[opp].length - 1);
                            }
                        }
                        else
                        {
                            // sum == 0, meaning opponent has never followed
                            // any of the available models.  Let's assume he follows
                            // the first available model which has never been
                            // previously active.  Note that this is extremely
                            // unlikely to happen, and if it does, it should
                            // be on one of the first rounds
                            m = 0;
                            while (m < uctBms[opp].length && (uctBms[opp][m].maxCount > 0.0 || !uctBms[opp][m].isAvailable()))
                            {
                                ++m;
                            }
                        }

                        if (m < uctBms[opp].length)
                        {
                            // Sample an action at random from the behaviour
                            int[] possibleOppActions = uctBms[opp][m].getActionList();
                            simulatedActions[turn][opp + 1] = possibleOppActions[rand.nextInt(possibleOppActions.length)];
                        }
                        else
                        {
                            // Ok, assume opponent acts randomly
                            simulatedActions[turn][opp + 1] = rand.nextInt(numActions);
                        }
                    }
                }
                else
                {
                    // Sample an action for each opponent
                    for (int opp = 0; opp < 2; ++opp)
                    {
                        // Calaculate opponent model's distribution over actions
                        double[] actionNums;
                        double[] actionDenoms;
                        double sum = 0.0;
                        if (turn == 0)
                        {
                            // Just grab the stored values
                            actionNums = nextActionNums[opp];
                            actionDenoms = nextActionDenoms[opp];
                            sum = nextActionProbsSum[opp];
                        }
                        else
                        {
                            actionNums = new double[numActions];
                            actionDenoms = new double[numActions];
                            for (int m = 0; m < uctBms[opp].length; ++m)
                            {
                                if (uctBms[opp][m].isAvailable() && uctBms[opp][m].maxCount > 0)
                                {
                                    int[] actions = uctBms[opp][m].getActionList();
                                    for (int a : actions)
                                    {
                                        actionNums[a] += uctBms[opp][m].count / actions.length;
                                        actionDenoms[a] += uctBms[opp][m].maxCount;
                                    }
                                }
                            }
                            for (int a = 0; a < numActions; ++a)
                            {
                                if (actionDenoms[a] > 0.0)
                                {
                                    sum += actionNums[a] / actionDenoms[a];
                                }
                            }
                        }
                        if (sum > 0)
                        {
                            double dart = rand.nextDouble() * sum;
                            double cumProb = 0.0;
                            int a;
                            for (a = 0; a < numActions; ++a)
                            {
                                if (actionDenoms[a] > 0)
                                {
                                    cumProb += actionNums[a] / actionDenoms[a];
                                    if (dart < cumProb)
                                    {
                                        break;
                                    }
                                }
                                assert(a < numActions - 1);
                            }
                            simulatedActions[turn][opp + 1] = a;
                        }
                        else
                        {
                            // sum == 0, meaning opponent has never followed
                            // any of the available models.  Let's assume he follows
                            // the first available model which has never been
                            // previously active
                            int m = 0;
                            while (m < uctBms[opp].length && (uctBms[opp][m].maxCount > 0.0 || !uctBms[opp][m].isAvailable()))
                            {
                                ++m;
                            }
                            if (m < uctBms[opp].length)
                            {
                                int[] possibleOppActions = uctBms[opp][m].getActionList();
                                simulatedActions[turn][opp + 1] = possibleOppActions[rand.nextInt(possibleOppActions.length)];
                            }
                            else
                            {
                                // Ok, assume opponent acts randomly
                                simulatedActions[turn][opp + 1] = rand.nextInt(numActions);
                            }
                        }
                    }
                }
 
                // Get the utility for the joint action
                util += utilFunc.getUtility(simulatedActions[turn])[0];

                // Update the UCT models
                for (int m = 0; m < uctBms[0].length; ++m)
                {
                    uctBms[0][m].update(simulatedActions[turn], false);
                    uctBms[1][m].update(simulatedActions[turn], false);
                }
            }

            // Back up the util value and increment number of visits
            for (int turn = 0; turn < maxDepth; ++turn)
            {
                int stateIndex = simulatedActions[turn][0];
                for (int i = 0; i < turn; ++i)
                {
                    stateIndex += (int)(Math.pow(numActions, i + 1)) * simulatedActions[i][0];
                }

                numVisits[turn][stateIndex]++;
                stateValues[turn][stateIndex] += (1.0 / numVisits[turn][stateIndex]) * (util - stateValues[turn][stateIndex]);
            }

            ++sim;
        }

        // Find the best looking action, randomly breaking ties.
        // Only move if the best action looks significantly better than our
        // previous action
        ArrayList<Integer> bestActions = new ArrayList();
        double bestActionValue = -Double.MAX_VALUE;
        for (int a = 0; a < numActions; ++a)
        {
            double actionValue = stateValues[0][a];

            if (actionValue > bestActionValue)
            {
                bestActions.clear();
                bestActions.add(a);
                bestActionValue = actionValue;
            }
            else if (actionValue == bestActionValue)
            {
                bestActions.add(a);
            }
        }

        int action = bestActions.get(rand.nextInt(bestActions.size()));
        if (bestActionValue < stateValues[0][prevActions[0]] + (MOVING_THRESHOLD * maxDepth))
        {
            action = prevActions[0];
        }

        walltimePassed += System.currentTimeMillis() - getActionStartTime;

        if (DO_DEBUG)
        {
            System.err.println("Time step = " + time);
            System.err.println("Num Simulations = " + sim);
            System.err.println("UCT action values:");
            for (int a = 0; a < numActions; ++a)
            {
                System.err.println(a + "=" + stateValues[0][a]);
            }
            System.err.println();
        }
        
        return action;
    }

    @Override
    public void observeOutcome(int[] actions)
    {
        long startTime = System.currentTimeMillis();

        // Update previous actions
        for (int p = 0; p < NUM_PLAYERS; ++p)
        {
            prevActions[p] = actions[p];
        }

        // Update the behaviour models
        for (int opp = 0; opp < NUM_PLAYERS - 1; ++opp)
        {
            for (int m = 0; m < bms[opp].size(); ++m)
            {
                bms[opp].get(m).update(actions, true);
            }
        }

        if (DO_SAMPLE_BEHAVIOURS)
        {
            // Optimization: Store the sum of the behaviour weights for the
            // next turn
            for (int opp = 0; opp < 2; ++opp)
            {
                nextBehaviourProbsSum[opp] = 0.0;
                for (int m = 0; m < bms[opp].size(); ++m)
                {
                    BehaviourModel bm = bms[opp].get(m);
                    if (bm.isAvailable())
                    {
                        nextBehaviourProbsSum[opp] += bm.getWeight();
                    }
                }
            }
        }
        else
        {
            // Optimization: Store the opponent model's distribution over actions
            // for the next turn
            for (int opp = 0; opp < 2; ++opp)
            {
                for (int a = 0; a < numActions; ++a)
                {
                    nextActionNums[opp][a] = 0.0;
                    nextActionDenoms[opp][a] = 0.0;
                }
                nextActionProbsSum[opp] = 0.0;
            }
            for (int opp = 0; opp < 2; ++opp)
            {
                for (int m = 0; m < bms[opp].size(); ++m)
                {
                    BehaviourModel bm = bms[opp].get(m);
                    if (bm.isAvailable() && bm.maxCount > 0)
                    {
                        int[] actionList = bm.getActionList();
                        for (int a : actionList)
                        {
                            nextActionNums[opp][a] += bm.count / actionList.length;
                            nextActionDenoms[opp][a] += bm.maxCount;
                        }
                    }
                }
                for (int a = 0; a < numActions; ++a)
                {
                    if (nextActionDenoms[opp][a] > 0.0)
                    {
                        nextActionProbsSum[opp] += nextActionNums[opp][a] / nextActionDenoms[opp][a];
                    }
                }
            }
        }

        ++time;

        walltimePassed += System.currentTimeMillis() - startTime;

        if (DO_DEBUG)
        {
            for (int opp = 0; opp < 2; ++opp)
            {
                System.err.println("\nUpdated Opponent " + (opp + 1) + " Models:");
                for (int m = 0; m < bms[opp].size(); ++m)
                {
                    BehaviourModel bm = bms[opp].get(m);
                    double truncatedCount = ((int)(bm.count * 100)) / 100.0;
                    double truncatedMaxCount = ((int)(bm.maxCount * 100)) / 100.0;
                    System.err.print(bm.getName() + ":\tavailable=" + bm.isAvailable() + "\tcount=" + truncatedCount + "\tmaxCount=" + truncatedMaxCount);
                    if (bm.isAvailable())
                    {
                        System.err.print("\tactionList=" + Arrays.toString(bm.getActionList()));
                    }
                    System.err.println("\tweight=" + bm.getWeight());
                }
                if (!DO_SAMPLE_BEHAVIOURS)
                {
                    System.err.print("Next action probs:");
                    for (int a = 0; a < numActions; ++a)
                    {
                        System.err.print(" " + a + "=");
                        if (nextActionDenoms[opp][a] > 0)
                        {
                            double prob = nextActionNums[opp][a] / (nextActionDenoms[opp][a] * nextActionProbsSum[opp]);
                            double truncatedProb = ((int)(prob * 100)) / 100.0;
                            if (prob - truncatedProb >= 0.005)
                            {
                                truncatedProb += 0.01;
                                truncatedProb = ((int)(truncatedProb * 100)) / 100.0;
                            }
                            System.err.print(truncatedProb);
                        }
                        else
                        {
                            System.err.print("0.00");
                        }
                    }
                }
                System.err.println();
            }
            System.err.println("\n");
        }
    }
}
