package main;

import game.ConstantGameGenerator;
import game.Game;
import game.GameGenerator;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Random;

import stats.SeriesObserver;
import stats.UtilitySeriesObserver;
import lemonade.ScaledLemonadeUtility;


/**
 * A main class for running the Lemonade tournament.
 * Every triple of players is played.
 * No variance reduction tricks are used, nor is variance calculated.
 * Usage: <SEED> <PLAYER1> <OPTION1> <PLAYER2> <OPTION2> <PLAYER3> <OPTION3>
 * @author maz
 *
 */
public class RunLemonadeTournament {

  
  public static SeriesObserver getObserver(GameGenerator g, String className) throws ClassNotFoundException, NoSuchMethodException, InstantiationException, InvocationTargetException, IllegalArgumentException, IllegalAccessException{
    String paramString = "(game.GameGenerator)";
    Class<?>[] params = new Class<?>[]{
        game.GameGenerator.class};
    try{
        Class<?> c = Class.forName(className);
        Constructor<?> constr = c.getConstructor(params);
        return (SeriesObserver)(constr.newInstance(g));
    } catch(ClassNotFoundException cnf){
      System.err.println("Could not find "+className);
      throw cnf;
    } catch(SecurityException s){
      System.err.println("No permission to access "+className+paramString);
      throw s;
    } catch (NoSuchMethodException nsm){
      System.err.println("Could not find "+className+paramString);
      throw nsm;
    } catch (InstantiationException ie){
      System.err.println("Cannot instantiate "+className);
      throw ie;
    } catch (InvocationTargetException ite){
      System.err.println("Constructor "+className+paramString+" threw an exception");
      throw ite;
    }
    
  }
  
  
  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void runTimeTest(GameGenerator g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      String[] threeStrategies = new String[]{"","strategy.UniformStrategy","strategy.UniformStrategy"};
      String[] threeOptions = new String[]{"","",""};
      threeStrategies[0]=strategies[i];
      threeOptions[0]=options[i];
      SeriesObserver observer = getObserver(g,observerClass);
      long initial = System.currentTimeMillis();
      RunLemonadeGame.manyRuns(g.clone(), threeStrategies, threeOptions, seedGenerator,iterations,repeats,observer);
      long finalTime = System.currentTimeMillis();
      double seconds = ((double)(finalTime-initial))/1000.0;
      double average = seconds/repeats;
      String passed = (average<10.0) ? "passed" : "failed";
      System.out.println(strategies[i]+" "+passed +" test. Time:"+average);

    }
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void runTournament(GameGenerator g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      for(int j=i+1;j<strategies.length;j++){
        for(int k=j+1;k<strategies.length;k++){
          String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
          String[] threeOptions = {options[i],options[j],options[k]};
          runGame(g, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observerClass);
        }
      }
    }
  }

  /**
   * Run a tournament with an arbitrary game.
   * Every three players (including repeats) are played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void runAsymmetricTournament(GameGenerator g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        for(int k=0;k<strategies.length;k++){
          String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
          String[] threeOptions = {options[i],options[j],options[k]};
          runGame(g, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observerClass);
        }
      }
    }
  }


  /**
   * Run a permuted tournament with an arbitrary game.
   * Every three players (not including repeats) are played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void runPermutedTournament(GameGenerator g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        for(int k=0;k<strategies.length;k++){
          if (i==k || i==j || j==k){
            continue;
          }
          String[] threeStrategies = {strategies[i],strategies[j],strategies[k]};
          String[] threeOptions = {options[i],options[j],options[k]};
          runGame(g, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observerClass);
        }
      }
    }
  }

  
  public static void runGame(GameGenerator g, String[] threeStrategies, String[] threeOptions, Random seedGenerator, int iterations, int repeats, String observerClass) throws IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, InstantiationException, InvocationTargetException, IllegalAccessException{
    GameGenerator g2 = g.clone();
    SeriesObserver observer = getObserver(g2,observerClass);
    RunLemonadeGame.manyRuns(g2, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observer);
    observer.display(threeStrategies);
    
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void run2vs1Tournament(GameGenerator g, String[] strategies, String[] options, Random seedGenerator,int iterations,int repeats, String observerClass) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    for(int i=0;i<strategies.length;i++){
      for(int j=0;j<strategies.length;j++){
        if (i!=j){
          String[] threeStrategies = {strategies[i],strategies[i],strategies[j]};
          String[] threeOptions = {options[i],options[i],options[j]};
          
          runGame(g, threeStrategies, threeOptions, seedGenerator,iterations,repeats,observerClass);
        }
      }
    }
  }

  public static void showUsage(){
    System.err.println("Usage:RunLemonadeTournament [-matchtype <MATCHTYPE>|-seed <SEED>|-actions <ACTIONS>| -rounds <ROUNDS> |-repeats <REPEATS>]* -players [<CLASS> <OPTION>]*");
    System.err.println("<MATCHTYPE>:= normal | normalpermuted | 2vs1 | all | asymmetric | timing");
    System.err.println("<SEED> a seed value for all the randomness in the system.");
    System.err.println("<ACTIONS> the number of actions, if one is using a traditional island model.");
    System.err.println("<ROUNDS> the number of rounds in a single game.");
    System.err.println("<REPEATS> the number of games in a series.");
    System.exit(0);
  }

  /**
   * Run a tournament with a symmetric game.
   * Every triple of players is played.
   * No variance reduction tricks are used, nor is variance calculated.
   */
  public static void main(String[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException{
    
    if (args.length<4){
      showUsage();
    }
    
    long seed=0;
    int numActions=12;
    int iterations = 100;
    int repeats = 1000;
    String observerClassName = UtilitySeriesObserver.class.getName();
    int playerArgIndex=args.length;
    String matchType = "normal";
    String generator = null;
    for(int i=0;i<args.length;i+=2){
      String arg = args[i];
      String fieldStr = args[i+1];
      if (arg.equals("-seed")){
        long field = Long.parseLong(fieldStr);
        seed = field;
      } else if (arg.equals("-observer")){
        observerClassName = fieldStr;
      } else if (arg.equals("-generator")){
        generator = fieldStr;
      } else if (arg.equals("-matchtype")){
        matchType = fieldStr;
        
      }else if (arg.equals("-actions")){
        long field = Long.parseLong(fieldStr);
        if (100L<field){
          System.err.println("Number of actions needs to be less than 100");
          showUsage();
        }
        numActions = (int)field;
      } else if (arg.equals("-rounds")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of rounds needs to be less than 2^31-1");
          showUsage();
        }
        iterations = (int)field;
      } else if (arg.equals("-repeats")){
        long field = Long.parseLong(fieldStr);
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of repeats needs to be less than 2^31-1");
          showUsage();
        }
        repeats = (int)field;
      } else if (arg.equals("-players")){
        playerArgIndex=i+1;
        break;
      } else {
        System.err.println("Unknown option:"+arg);
        showUsage();
        
      }
    }

    Random seedGenerator = new Random(seed);
    ScaledLemonadeUtility utility = new ScaledLemonadeUtility(numActions);
    GameGenerator g = new ConstantGameGenerator(new Game(new int[]{numActions,numActions,numActions},utility));

    if (generator!=null){
      long generatorSeed = seedGenerator.nextLong();
      Class<?> generatorClass = Class.forName(generator);
      Constructor<?> constructor = generatorClass.getConstructor(long.class);
      g = (GameGenerator)constructor.newInstance(generatorSeed);      
    }
    String[] strategies = new String[(args.length-playerArgIndex)/2];
    String[] options = new String[strategies.length];
    int playerIndex = 0;
    for(int i=playerArgIndex;i<args.length;i+=2){
      strategies[playerIndex]=args[i];
      options[playerIndex]=args[i+1];
      playerIndex++;
    }
    
    if (matchType.equals("normal")){
      runTournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    } else if (matchType.equals("normalpermuted")){
      runPermutedTournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    } else if (matchType.equals("2vs1")){
      run2vs1Tournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
      
    } else if (matchType.equals("all")){
      runTournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
      run2vs1Tournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    } else if (matchType.equals("asymmetric")){
      runAsymmetricTournament(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    } else if (matchType.equals("timing")){
      runTimeTest(g, strategies, options, seedGenerator,iterations,repeats,observerClassName);
    }
  }
}
